@@ -31,6 +31,9 @@ for my $text (@text)
 
 	$result = $parser -> parse(\$text);
 
+	print join("\n", @{$parser -> tree2string}), "\n";
+	print "Parse result: $result (0 is success)\n";
+
 	if ($count == 3)
 	{
 		print "Deliberate error: Failed to parse |$text|\n";
